classdef CsfWriter < handle
  % CsfWriter
  %   The csf writer class takes care of writing depth frames to disk.  The
  %   class has to be created from either a Camera or csf reader, because it
  %   needs information about the camera that was used to record the depth
  %   frames.  Therefore you should only write depth frames coming from one
  %   camera into the same writer

  %
  % Methods:
  %   writeFrame - Write frame to csf file
  %

  properties(Access=protected)
    mPtr = uint64(0);
  end

  methods(Access=public)
    function obj = CsfWriter(varargin)
      % CsfWriter
      %   Construct new csf_writer class
      %
      % ----------
      %
      %   Construct csf writer from calibration and configuration
      %
      % Inputs:
      %   1 - string - Path to where the csf file should be created
      %   2 - CameraConfig - Camera configuration
      %   3 - Calibration - Calibration
      %   4 - string - Serial number of camera
      %
      % ----------
      %
      %   Construct the csf writer from a camera
      %
      % Inputs:
      %   1 - string - Path to where the csf file should be created
      %   2 - Camera - The camera to get the camera config and calibration file
      %                from
      %

      if nargin == 1 && isa(varargin{1}, 'uint64')
        obj.mPtr = varargin{1};
      elseif nargin == 4 && (isStringScalar(varargin{1}) || ischar(varargin{1})) && isa(varargin{2}, 'tof.CameraConfig') && isa(varargin{3}, 'tof.Calibration') && (isStringScalar(varargin{4}) || ischar(varargin{4}))
        obj.mPtr = tof.mex('tof_csf_writer_new', varargin{1}, varargin{2}.get_object_ptr(), varargin{3}.get_object_ptr(), varargin{4});
      elseif nargin == 2 && (isStringScalar(varargin{1}) || ischar(varargin{1})) && isa(varargin{2}, 'tof.Camera')
        obj.mPtr = tof.mex('tof_csf_writer_new_from_camera', varargin{1}, varargin{2}.get_object_ptr());
      else
        error("Can't construct empty object");
      end
    end


    function delete(obj)
      if obj.mPtr ~= 0
        tof.mex('tof_csf_writer_delete', obj.mPtr);
        obj.mPtr = 0;
      end
    end

    function ptr = get_object_ptr(obj)
      ptr = obj.mPtr;
    end

    function writeFrame(obj, frame)
      % writeFrame
      %   Write frame to csf file
      %
      % Inputs:
      %   frame - Data - A data frame
      %

      tof.mex('tof_csf_writer_write_frame', obj.mPtr, frame.get_object_ptr());
    end

  end
end
